@@ -12,8 +12,6 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import string
-import random
 import time
 import logging
 import json
@@ -24,8 +22,8 @@ import tempfile
 import warnings
 import yaml
 import zipfile
-from datetime import datetime
-from typing import Mapping, Callable
+import datetime
+from typing import Mapping, Callable, Optional
 
 import kfp
 import kfp_server_api
@@ -35,12 +33,26 @@ from kfp.compiler._k8s_helper import sanitize_k8s_name
 
 from kfp._auth import get_auth_token, get_gcp_access_token
 
-
+# TTL of the access token associated with the client. This is needed because
+# `gcloud auth print-access-token` generates a token with TTL=1 hour, after
+# which the authentication expires. This TTL is needed for kfp.Client()
+# initialized with host=<inverse proxy endpoint>.
+# Set to 55 mins to provide some safe margin.
+_GCP_ACCESS_TOKEN_TIMEOUT = datetime.timedelta(minutes=55)
+# Operators on scalar values. Only applies to one of |int_value|,
+# |long_value|, |string_value| or |timestamp_value|.
+_FILTER_OPERATIONS = {"UNKNOWN": 0,
+    "EQUALS" : 1,
+    "NOT_EQUALS" : 2,
+    "GREATER_THAN": 3,
+    "GREATER_THAN_EQUALS": 5,
+    "LESS_THAN": 6,
+    "LESS_THAN_EQUALS": 7}
 
 def _add_generated_apis(target_struct, api_module, api_client):
-  '''Initializes a hierarchical API object based on the generated API module.
+  """Initializes a hierarchical API object based on the generated API module.
   PipelineServiceApi.create_pipeline becomes target_struct.pipelines.create_pipeline
-  '''
+  """
   Struct = type('Struct', (), {})
 
   def camel_case_to_snake_case(name):
@@ -76,55 +88,86 @@ KF_PIPELINES_OVERRIDE_EXPERIMENT_NAME = 'KF_PIPELINES_OVERRIDE_EXPERIMENT_NAME'
 
 
 class Client(object):
-  """ API Client for KubeFlow Pipeline.
+  """API Client for KubeFlow Pipeline.
+
+  Args:
+    host: The host name to use to talk to Kubeflow Pipelines. If not set, the in-cluster
+        service DNS name will be used, which only works if the current environment is a pod
+        in the same cluster (such as a Jupyter instance spawned by Kubeflow's
+        JupyterHub). If you have a different connection to cluster, such as a kubectl
+        proxy connection, then set it to something like "127.0.0.1:8080/pipeline.
+        If you connect to an IAP enabled cluster, set it to
+        https://<your-deployment>.endpoints.<your-project>.cloud.goog/pipeline".
+    client_id: The client ID used by Identity-Aware Proxy.
+    namespace: The namespace where the kubeflow pipeline system is run.
+    other_client_id: The client ID used to obtain the auth codes and refresh tokens.
+        Reference: https://cloud.google.com/iap/docs/authentication-howto#authenticating_from_a_desktop_app.
+    other_client_secret: The client secret used to obtain the auth codes and refresh tokens.
+    existing_token: Pass in token directly, it's used for cases better get token outside of SDK, e.x. GCP Cloud Functions
+        or caller already has a token
+    cookies: CookieJar object containing cookies that will be passed to the pipelines API.
+    proxy: HTTP or HTTPS proxy server
+    ssl_ca_cert: Cert for proxy
+    kube_context: String name of context within kubeconfig to use, defaults to the current-context set within kubeconfig.
   """
 
   # in-cluster DNS name of the pipeline service
   IN_CLUSTER_DNS_NAME = 'ml-pipeline.{}.svc.cluster.local:8888'
   KUBE_PROXY_PATH = 'api/v1/namespaces/{}/services/ml-pipeline:http/proxy/'
 
+  # Auto populated path in pods
+  # https://kubernetes.io/docs/tasks/access-application-cluster/access-cluster/#accessing-the-api-from-a-pod
+  # https://kubernetes.io/docs/reference/access-authn-authz/service-accounts-admin/#serviceaccount-admission-controller
+  NAMESPACE_PATH = '/var/run/secrets/kubernetes.io/serviceaccount/namespace'
+
   LOCAL_KFP_CONTEXT = os.path.expanduser('~/.config/kfp/context.json')
 
   # TODO: Wrap the configurations for different authentication methods.
-  def __init__(self, host=None, client_id=None, namespace='kubeflow', other_client_id=None, other_client_secret=None, existing_token=None):
+  def __init__(self, host=None, client_id=None, namespace='kubeflow', other_client_id=None, other_client_secret=None, existing_token=None, cookies=None, proxy=None, ssl_ca_cert=None, kube_context=None):
     """Create a new instance of kfp client.
-
-    Args:
-      host: the host name to use to talk to Kubeflow Pipelines. If not set, the in-cluster
-          service DNS name will be used, which only works if the current environment is a pod
-          in the same cluster (such as a Jupyter instance spawned by Kubeflow's
-          JupyterHub). If you have a different connection to cluster, such as a kubectl
-          proxy connection, then set it to something like "127.0.0.1:8080/pipeline.
-          If you connect to an IAP enabled cluster, set it to
-          https://<your-deployment>.endpoints.<your-project>.cloud.goog/pipeline".
-      client_id: The client ID used by Identity-Aware Proxy.
-      namespace: the namespace where the kubeflow pipeline system is run.
-      other_client_id: The client ID used to obtain the auth codes and refresh tokens.
-          Reference: https://cloud.google.com/iap/docs/authentication-howto#authenticating_from_a_desktop_app.
-      other_client_secret: The client secret used to obtain the auth codes and refresh tokens.
-      existing_token: pass in token directly, it's used for cases better get token outside of SDK, e.x. GCP Cloud Functions
-          or caller already has a token
     """
     host = host or os.environ.get(KF_PIPELINES_ENDPOINT_ENV)
     self._uihost = os.environ.get(KF_PIPELINES_UI_ENDPOINT_ENV, host)
-    config = self._load_config(host, client_id, namespace, other_client_id, other_client_secret, existing_token)
-    api_client = kfp_server_api.api_client.ApiClient(config)
+    config = self._load_config(host, client_id, namespace, other_client_id, other_client_secret, existing_token, proxy, ssl_ca_cert, kube_context)
+    # Save the loaded API client configuration, as a reference if update is
+    # needed.
+    self._load_context_setting_or_default()
+    self._existing_config = config
+    if cookies is None:
+      cookies = self._context_setting.get('client_authentication_cookie')
+    api_client = kfp_server_api.api_client.ApiClient(config, cookie=cookies,
+        header_name=self._context_setting.get('client_authentication_header_name'),
+        header_value=self._context_setting.get('client_authentication_header_value'))
     _add_generated_apis(self, kfp_server_api, api_client)
     self._job_api = kfp_server_api.api.job_service_api.JobServiceApi(api_client)
     self._run_api = kfp_server_api.api.run_service_api.RunServiceApi(api_client)
     self._experiment_api = kfp_server_api.api.experiment_service_api.ExperimentServiceApi(api_client)
     self._pipelines_api = kfp_server_api.api.pipeline_service_api.PipelineServiceApi(api_client)
     self._upload_api = kfp_server_api.api.PipelineUploadServiceApi(api_client)
-    self._load_context_setting_or_default()
+    self._healthz_api = kfp_server_api.api.healthz_service_api.HealthzServiceApi(api_client)
+    if not self._context_setting['namespace'] and self.get_kfp_healthz().multi_user is True:
+      try:
+        with open(Client.NAMESPACE_PATH, 'r') as f:
+          current_namespace = f.read()
+          self.set_user_namespace(current_namespace)
+      except FileNotFoundError:
+        logging.info('Failed to automatically set namespace.', exc_info=True)
 
-  def _load_config(self, host, client_id, namespace, other_client_id, other_client_secret, existing_token):
+  def _load_config(self, host, client_id, namespace, other_client_id, other_client_secret, existing_token, proxy, ssl_ca_cert, kube_context):
     config = kfp_server_api.configuration.Configuration()
 
+    if proxy:
+      # https://github.com/kubeflow/pipelines/blob/c6ac5e0b1fd991e19e96419f0f508ec0a4217c29/backend/api/python_http_client/kfp_server_api/rest.py#L100
+      config.proxy = proxy
+
+    if ssl_ca_cert:
+      config.ssl_ca_cert = ssl_ca_cert
+
     host = host or ''
     # Preprocess the host endpoint to prevent some common user mistakes.
-    # This should only be done for non-IAP cases (when client_id is None). IAP requires preserving the protocol.
     if not client_id:
-      host = re.sub(r'^(http|https)://', '', host).rstrip('/')
+      # always preserving the protocol (http://localhost requires it)
+      host = host.rstrip('/')
 
     if host:
       config.host = host
@@ -150,10 +193,13 @@ class Client(object):
     #
     if existing_token:
       token = existing_token
+      self._is_refresh_token = False
     elif client_id:
       token = get_auth_token(client_id, other_client_id, other_client_secret)
+      self._is_refresh_token = True
     elif self._is_inverse_proxy_host(host):
       token = get_gcp_access_token()
+      self._is_refresh_token = False
 
     if token:
       config.api_key['authorization'] = token
@@ -177,7 +223,7 @@ class Client(object):
       return config
 
     try:
-      k8s.config.load_kube_config(client_configuration=config)
+      k8s.config.load_kube_config(client_configuration=config, context=kube_context)
     except:
       print('Failed to load kube config.')
       return config
@@ -226,19 +272,52 @@ class Client(object):
       self._context_setting = {
         'namespace': '',
       }
+      
+  def _refresh_api_client_token(self):
+    """Refreshes the existing token associated with the kfp_api_client."""
+    if getattr(self, '_is_refresh_token', None):
+      return
+
+    new_token = get_gcp_access_token()
+    self._existing_config.api_key['authorization'] = new_token
 
   def set_user_namespace(self, namespace):
     """Set user namespace into local context setting file.
-       This function should only be used when Kubeflow Pipelines is in the multi-user mode.
+    
+    This function should only be used when Kubeflow Pipelines is in the multi-user mode.
+
     Args:
       namespace: kubernetes namespace the user has access to.
     """
     self._context_setting['namespace'] = namespace
+    if not os.path.exists(os.path.dirname(Client.LOCAL_KFP_CONTEXT)):
+        os.makedirs(os.path.dirname(Client.LOCAL_KFP_CONTEXT))
     with open(Client.LOCAL_KFP_CONTEXT, 'w') as f:
       json.dump(self._context_setting, f)
 
+  def get_kfp_healthz(self):
+    """Gets healthz info of KFP deployment.
+
+    Returns:
+      response: json formatted response from the healtz endpoint.
+    """
+    count = 0
+    response = None
+    max_attempts = 5
+    while not response:
+      count += 1
+      if count > max_attempts:
+        raise TimeoutError('API call timeout')
+      try:
+        response = self._healthz_api.get_healthz()
+        return response
+      except:
+        logging.info('Failed to get healthz info attempt {} of 5.'.format(count))
+        time.sleep(5)
+
   def get_user_namespace(self):
     """Get user namespace in context config.
+
     Returns:
       namespace: kubernetes namespace from the local context file or empty if it wasn't set.
     """
@@ -246,12 +325,14 @@ class Client(object):
 
   def create_experiment(self, name, description=None, namespace=None):
     """Create a new experiment.
+
     Args:
-      name: the name of the experiment.
-      description: description of the experiment.
-      namespace: kubernetes namespace where the experiment should be created.
+      name: The name of the experiment.
+      description: Description of the experiment.
+      namespace: Kubernetes namespace where the experiment should be created.
         For single user deployment, leave it as None;
         For multi user, input a namespace where the user is authorized.
+
     Returns:
       An Experiment object. Most important field is id.
     """
@@ -259,9 +340,10 @@ class Client(object):
     experiment = None
     try:
       experiment = self.get_experiment(experiment_name=name, namespace=namespace)
-    except:
+    except ValueError as error:
       # Ignore error if the experiment does not exist.
-      pass
+      if not str(error).startswith('No experiment is found with name'):
+        raise error
 
     if not experiment:
       logging.info('Creating experiment {}.'.format(name))
@@ -281,20 +363,49 @@ class Client(object):
     if self._is_ipython():
       import IPython
       html = \
-          ('Experiment link <a href="%s/#/experiments/details/%s" target="_blank" >here</a>'
+          ('<a href="%s/#/experiments/details/%s" target="_blank" >Experiment details</a>.'
           % (self._get_url_prefix(), experiment.id))
       IPython.display.display(IPython.display.HTML(html))
     return experiment
 
+  def get_pipeline_id(self, name):
+    """Find the id of a pipeline by name.
+
+    Args:
+      name: Pipeline name.
+
+    Returns:
+      Returns the pipeline id if a pipeline with the name exists.
+    """
+    pipeline_filter = json.dumps({
+      "predicates": [
+        {
+          "op":  _FILTER_OPERATIONS["EQUALS"],
+          "key": "name",
+          "stringValue": name,
+        }
+      ]
+    })
+    result = self._pipelines_api.list_pipelines(filter=pipeline_filter)
+    if result.pipelines is None:
+      return None
+    if len(result.pipelines)==1:
+      return result.pipelines[0].id
+    elif len(result.pipelines)>1:
+      raise ValueError("Multiple pipelines with the name: {} found, the name needs to be unique".format(name))
+    return None
+
   def list_experiments(self, page_token='', page_size=10, sort_by='', namespace=None):
     """List experiments.
+
     Args:
-      page_token: token for starting of the page.
-      page_size: size of the page.
-      sort_by: can be '[field_name]', '[field_name] des'. For example, 'name des'.
-      namespace: kubernetes namespace where the experiment was created.
+      page_token: Token for starting of the page.
+      page_size: Size of the page.
+      sort_by: Can be '[field_name]', '[field_name] des'. For example, 'name desc'.
+      namespace: Kubernetes namespace where the experiment was created.
         For single user deployment, leave it as None;
         For multi user, input a namespace where the user is authorized.
+  
     Returns:
       A response object including a list of experiments and next page token.
     """
@@ -309,15 +420,19 @@ class Client(object):
 
   def get_experiment(self, experiment_id=None, experiment_name=None, namespace=None):
     """Get details of an experiment
+
     Either experiment_id or experiment_name is required
+
     Args:
-      experiment_id: id of the experiment. (Optional)
-      experiment_name: name of the experiment. (Optional)
-      namespace: kubernetes namespace where the experiment was created.
+      experiment_id: Id of the experiment. (Optional)
+      experiment_name: Name of the experiment. (Optional)
+      namespace: Kubernetes namespace where the experiment was created.
         For single user deployment, leave it as None;
         For multi user, input the namespace where the user is authorized.
+
     Returns:
       A response object including details of a experiment.
+
     Throws:
       Exception if experiment is not found or None of the arguments is provided
     """
@@ -330,7 +445,7 @@ class Client(object):
     while next_page_token is not None:
       list_experiments_response = self.list_experiments(page_size=100, page_token=next_page_token, namespace=namespace)
       next_page_token = list_experiments_response.next_page_token
-      for experiment in list_experiments_response.experiments:
+      for experiment in list_experiments_response.experiments or []:
         if experiment.name == experiment_name:
           return self._experiment_api.get_experiment(id=experiment.id)
     raise ValueError('No experiment is found with name {}.'.format(experiment_name))
@@ -363,30 +478,52 @@ class Client(object):
       with open(package_file, 'r') as f:
         return yaml.safe_load(f)
     else:
-      raise ValueError('The package_file '+ package_file + ' should ends with one of the following formats: [.tar.gz, .tgz, .zip, .yaml, .yml]')
+      raise ValueError('The package_file '+ package_file + ' should end with one of the following formats: [.tar.gz, .tgz, .zip, .yaml, .yml]')
 
   def list_pipelines(self, page_token='', page_size=10, sort_by=''):
     """List pipelines.
+
     Args:
-      page_token: token for starting of the page.
-      page_size: size of the page.
-      sort_by: one of 'field_name', 'field_name des'. For example, 'name des'.
+      page_token: Token for starting of the page.
+      page_size: Size of the page.
+      sort_by: one of 'field_name', 'field_name desc'. For example, 'name desc'.
+
     Returns:
       A response object including a list of pipelines and next page token.
     """
     return self._pipelines_api.list_pipelines(page_token=page_token, page_size=page_size, sort_by=sort_by)
 
+  def list_pipeline_versions(self, pipeline_id: str, page_token='', page_size=10, sort_by=''):
+    """List all versions of a given pipeline.
+
+    Args:
+      pipeline_id: The id of a pipeline.
+      page_token: Token for starting of the page.
+      page_size: Size of the page.
+        sort_by: one of 'field_name', 'field_name desc'. For example, 'name desc'.
+
+    Returns:
+      A response object including a list of pipeline versions and next page token.
+    """
+    return self._pipelines_api.list_pipeline_versions(
+        resource_key_type="PIPELINE",
+        resource_key_id=pipeline_id,
+        page_token=page_token,
+        page_size=page_size,
+        sort_by=sort_by
+    )
+
   # TODO: provide default namespace, similar to kubectl default namespaces.
   def run_pipeline(self, experiment_id, job_name, pipeline_package_path=None, params={}, pipeline_id=None, version_id=None):
     """Run a specified pipeline.
 
     Args:
-      experiment_id: The string id of an experiment.
-      job_name: name of the job.
-      pipeline_package_path: local path of the pipeline package(the filename should end with one of the following .tar.gz, .tgz, .zip, .yaml, .yml).
-      params: a dictionary with key (string) as param name and value (string) as as param value.
-      pipeline_id: the string ID of a pipeline.
-      version_id: the string ID of a pipeline version.
+      experiment_id: The id of an experiment.
+      job_name: Name of the job.
+      pipeline_package_path: Local path of the pipeline package(the filename should end with one of the following .tar.gz, .tgz, .zip, .yaml, .yml).
+      params: A dictionary with key (string) as param name and value (string) as as param value.
+      pipeline_id: The id of a pipeline.
+      version_id: The id of a pipeline version.
         If both pipeline_id and version_id are specified, version_id will take precendence.
         If only pipeline_id is specified, the default version of this pipeline is used to create the run.
 
@@ -406,16 +543,17 @@ class Client(object):
 
     if self._is_ipython():
       import IPython
-      html = ('Run link <a href="%s/#/runs/details/%s" target="_blank" >here</a>'
+      html = ('<a href="%s/#/runs/details/%s" target="_blank" >Run details</a>.'
               % (self._get_url_prefix(), response.run.id))
       IPython.display.display(IPython.display.HTML(html))
     return response.run
 
   def create_recurring_run(self, experiment_id, job_name, description=None, start_time=None, end_time=None, interval_second=None, cron_expression=None, max_concurrency=1, no_catchup=None, params={}, pipeline_package_path=None, pipeline_id=None, version_id=None, enabled=True):
     """Create a recurring run.
+
     Args:
       experiment_id: The string id of an experiment.
-      job_name: name of the job.
+      job_name: Name of the job.
       description: An optional job description.
       start_time: The RFC3339 time string of the time when to start the job.
       end_time: The RFC3339 time string of the time when to end the job.
@@ -436,6 +574,7 @@ class Client(object):
         If both pipeline_id and version_id are specified, pipeline_id will take precendence
         This will change in a future version, so it is recommended to use version_id by itself.
       enabled: A bool indicating whether the recurring run is enabled or disabled.
+
     Returns:
       A Job object. Most important field is id.
     """
@@ -472,14 +611,16 @@ class Client(object):
 
   def _create_job_config(self, experiment_id, params, pipeline_package_path, pipeline_id, version_id):
     """Create a JobConfig with spec and resource_references.
+
     Args:
-      experiment_id: The string id of an experiment.
+      experiment_id: The id of an experiment.
       pipeline_package_path: Local path of the pipeline package(the filename should end with one of the following .tar.gz, .tgz, .zip, .yaml, .yml).
       params: A dictionary with key (string) as param name and value (string) as param value.
-      pipeline_id: The string ID of a pipeline.
-      version_id: The string ID of a pipeline version. 
+      pipeline_id: The id of a pipeline.
+      version_id: The id of a pipeline version. 
         If both pipeline_id and version_id are specified, pipeline_id will take precendence
         This will change in a future version, so it is recommended to use version_id by itself.
+
     Returns:
       A JobConfig object with attributes spec and resource_reference.
     """
@@ -516,8 +657,16 @@ class Client(object):
         parameters=api_params)
     return JobConfig(spec=spec, resource_references=resource_references)
 
-  def create_run_from_pipeline_func(self, pipeline_func: Callable, arguments: Mapping[str, str], run_name=None, experiment_name=None, pipeline_conf: kfp.dsl.PipelineConf = None, namespace=None):
-    '''Runs pipeline on KFP-enabled Kubernetes cluster.
+  def create_run_from_pipeline_func(
+      self,
+      pipeline_func: Callable,
+      arguments: Mapping[str, str],
+      run_name: Optional[str] = None,
+      experiment_name: Optional[str] = None,
+      pipeline_conf: Optional[kfp.dsl.PipelineConf] = None,
+      namespace: Optional[str] = None):
+    """Runs pipeline on KFP-enabled Kubernetes cluster.
+
     This command compiles the pipeline function, creates or gets an experiment and submits the pipeline for execution.
 
     Args:
@@ -525,33 +674,41 @@ class Client(object):
       arguments: Arguments to the pipeline function provided as a dict.
       run_name: Optional. Name of the run to be shown in the UI.
       experiment_name: Optional. Name of the experiment to add the run to.
-      namespace: kubernetes namespace where the pipeline runs are created.
+      pipeline_conf: Optional. Pipeline configuration ops that will be applied
+        to all the ops in the pipeline func.
+      namespace: Kubernetes namespace where the pipeline runs are created.
         For single user deployment, leave it as None;
         For multi user, input a namespace where the user is authorized
-    '''
+    """
     #TODO: Check arguments against the pipeline function
     pipeline_name = pipeline_func.__name__
-    run_name = run_name or pipeline_name + ' ' + datetime.now().strftime('%Y-%m-%d %H-%M-%S')
-    try:
-      (_, pipeline_package_path) = tempfile.mkstemp(suffix='.zip')
+    run_name = run_name or pipeline_name + ' ' + datetime.datetime.now().strftime('%Y-%m-%d %H-%M-%S')
+    with tempfile.TemporaryDirectory() as tmpdir:
+      pipeline_package_path = os.path.join(tmpdir, 'pipeline.yaml')
       compiler.Compiler().compile(pipeline_func, pipeline_package_path, pipeline_conf=pipeline_conf)
       return self.create_run_from_pipeline_package(pipeline_package_path, arguments, run_name, experiment_name, namespace)
-    finally:
-      os.remove(pipeline_package_path)
 
-  def create_run_from_pipeline_package(self, pipeline_file: str, arguments: Mapping[str, str], run_name=None, experiment_name=None, namespace=None):
-    '''Runs pipeline on KFP-enabled Kubernetes cluster.
-    This command compiles the pipeline function, creates or gets an experiment and submits the pipeline for execution.
+  def create_run_from_pipeline_package(
+      self,
+      pipeline_file: str,
+      arguments: Mapping[str, str],
+      run_name: Optional[str] = None,
+      experiment_name: Optional[str] = None,
+      namespace: Optional[str] = None):
+    """Runs pipeline on KFP-enabled Kubernetes cluster.
+
+    This command takes a local pipeline package, creates or gets an experiment
+    and submits the pipeline for execution.
 
     Args:
       pipeline_file: A compiled pipeline package file.
       arguments: Arguments to the pipeline function provided as a dict.
       run_name: Optional. Name of the run to be shown in the UI.
       experiment_name: Optional. Name of the experiment to add the run to.
-      namespace: kubernetes namespace where the pipeline runs are created.
+      namespace: Kubernetes namespace where the pipeline runs are created.
         For single user deployment, leave it as None;
         For multi user, input a namespace where the user is authorized
-    '''
+    """
 
     class RunPipelineResult:
       def __init__(self, client, run_info):
@@ -560,7 +717,7 @@ class Client(object):
         self.run_id = run_info.id
 
       def wait_for_run_completion(self, timeout=None):
-        timeout = timeout or datetime.datetime.max - datetime.datetime.min
+        timeout = timeout or datetime.timedelta.max
         return self._client.wait_for_run_completion(self.run_id, timeout)
 
       def __repr__(self):
@@ -574,21 +731,25 @@ class Client(object):
       import warnings
       warnings.warn('Changing experiment name from "{}" to "{}".'.format(experiment_name, overridden_experiment_name))
     experiment_name = overridden_experiment_name or 'Default'
-    run_name = run_name or pipeline_name + ' ' + datetime.now().strftime('%Y-%m-%d %H-%M-%S')
+    run_name = run_name or (pipeline_name + ' ' +
+                            datetime.datetime.now().strftime(
+                                '%Y-%m-%d %H-%M-%S'))
     experiment = self.create_experiment(name=experiment_name, namespace=namespace)
     run_info = self.run_pipeline(experiment.id, run_name, pipeline_file, arguments)
     return RunPipelineResult(self, run_info)
 
   def list_runs(self, page_token='', page_size=10, sort_by='', experiment_id=None, namespace=None):
-    """List runs.
+    """List runs, optionally can be filtered by experiment or namespace.
+
     Args:
-      page_token: token for starting of the page.
-      page_size: size of the page.
-      sort_by: one of 'field_name', 'field_name des'. For example, 'name des'.
-      experiment_id: experiment id to filter upon
-      namespace: kubernetes namespace to filter upon.
+      page_token: Token for starting of the page.
+      page_size: Size of the page.
+      sort_by: One of 'field_name', 'field_name desc'. For example, 'name desc'.
+      experiment_id: Experiment id to filter upon
+      namespace: Kubernetes namespace to filter upon.
         For single user deployment, leave it as None;
         For multi user, input a namespace where the user is authorized.
+
     Returns:
       A response object including a list of experiments and next page token.
     """
@@ -603,11 +764,13 @@ class Client(object):
 
   def list_recurring_runs(self, page_token='', page_size=10, sort_by='', experiment_id=None):
     """List recurring runs.
+
     Args:
-      page_token: token for starting of the page.
-      page_size: size of the page.
-      sort_by: one of 'field_name', 'field_name des'. For example, 'name des'.
-      experiment_id: experiment id to filter upon
+      page_token: Token for starting of the page.
+      page_size: Size of the page.
+      sort_by: One of 'field_name', 'field_name desc'. For example, 'name desc'.
+      experiment_id: Experiment id to filter upon.
+
     Returns:
       A response object including a list of recurring_runs and next page token.
     """
@@ -619,10 +782,13 @@ class Client(object):
 
   def get_recurring_run(self, job_id):
     """Get recurring_run details.
+
     Args:
-      id of the recurring_run.
+      job_id: id of the recurring_run.
+
     Returns:
       A response object including details of a recurring_run.
+
     Throws:
       Exception if recurring_run is not found.
     """
@@ -631,29 +797,47 @@ class Client(object):
 
   def get_run(self, run_id):
     """Get run details.
+
     Args:
-      id of the run.
+      run_id: id of the run.
+
     Returns:
       A response object including details of a run.
+
     Throws:
       Exception if run is not found.
     """
     return self._run_api.get_run(run_id=run_id)
 
   def wait_for_run_completion(self, run_id, timeout):
-    """Wait for a run to complete.
+    """Waits for a run to complete.
+
     Args:
-      run_id: run id, returned from run_pipeline.
-      timeout: timeout in seconds.
+      run_id: Run id, returned from run_pipeline.
+      timeout: Timeout in seconds.
+
     Returns:
-      A run detail object: Most important fields are run and pipeline_runtime
+      A run detail object: Most important fields are run and pipeline_runtime.
+
+    Raises:
+      TimeoutError: if the pipeline run failed to finish before the specified timeout.
     """
     status = 'Running:'
-    start_time = datetime.now()
-    while status is None or status.lower() not in ['succeeded', 'failed', 'skipped', 'error']:
+    start_time = datetime.datetime.now()
+    last_token_refresh_time = datetime.datetime.now()
+    if isinstance(timeout, datetime.timedelta):
+      timeout = timeout.total_seconds()
+    while (status is None or
+           status.lower() not in ['succeeded', 'failed', 'skipped', 'error']):
+      # Refreshes the access token before it hits the TTL.
+      if (datetime.datetime.now() - last_token_refresh_time
+          > _GCP_ACCESS_TOKEN_TIMEOUT):
+        self._refresh_api_client_token()
+        last_token_refresh_time = datetime.datetime.now()
+        
       get_run_response = self._run_api.get_run(run_id=run_id)
       status = get_run_response.run.status
-      elapsed_time = (datetime.now() - start_time).seconds
+      elapsed_time = (datetime.datetime.now() - start_time).total_seconds()
       logging.info('Waiting for the job to complete...')
       if elapsed_time > timeout:
         raise TimeoutError('Run timeout')
@@ -662,38 +846,89 @@ class Client(object):
 
   def _get_workflow_json(self, run_id):
     """Get the workflow json.
+
     Args:
       run_id: run id, returned from run_pipeline.
+
     Returns:
-      workflow: json workflow
+      workflow: Json workflow
     """
     get_run_response = self._run_api.get_run(run_id=run_id)
     workflow = get_run_response.pipeline_runtime.workflow_manifest
     workflow_json = json.loads(workflow)
     return workflow_json
 
-  def upload_pipeline(self, pipeline_package_path, pipeline_name=None):
+  def upload_pipeline(
+    self,
+    pipeline_package_path: str = None,
+    pipeline_name: str = None,
+    description: str = None,
+  ):
     """Uploads the pipeline to the Kubeflow Pipelines cluster.
+
     Args:
       pipeline_package_path: Local path to the pipeline package.
       pipeline_name: Optional. Name of the pipeline to be shown in the UI.
+      description: Optional. Description of the pipeline to be shown in the UI.
+
     Returns:
       Server response object containing pipleine id and other information.
     """
 
-    response = self._upload_api.upload_pipeline(pipeline_package_path, name=pipeline_name)
+    response = self._upload_api.upload_pipeline(pipeline_package_path, name=pipeline_name, description=description)
     if self._is_ipython():
       import IPython
-      html = 'Pipeline link <a href=%s/#/pipelines/details/%s>here</a>' % (self._get_url_prefix(), response.id)
+      html = '<a href=%s/#/pipelines/details/%s>Pipeline details</a>.' % (self._get_url_prefix(), response.id)
+      IPython.display.display(IPython.display.HTML(html))
+    return response
+
+  def upload_pipeline_version(
+    self,
+    pipeline_package_path,
+    pipeline_version_name: str,
+    pipeline_id: Optional[str] = None,
+    pipeline_name: Optional[str] = None
+  ):
+    """Uploads a new version of the pipeline to the Kubeflow Pipelines cluster.
+    Args:
+      pipeline_package_path: Local path to the pipeline package.
+      pipeline_version_name:  Name of the pipeline version to be shown in the UI.
+      pipeline_id: Optional. Id of the pipeline.
+      pipeline_name: Optional. Name of the pipeline.
+    Returns:
+      Server response object containing pipleine id and other information.
+    Throws:
+      ValueError when none or both of pipeline_id or pipeline_name are specified
+      Exception if pipeline id is not found.
+    """
+
+    if all([pipeline_id, pipeline_name]) or not any([pipeline_id, pipeline_name]):
+      raise ValueError('Either pipeline_id or pipeline_name is required')
+
+    if pipeline_name:
+      pipeline_id = self.get_pipeline_id(pipeline_name)
+
+    response = self._upload_api.upload_pipeline_version(
+      pipeline_package_path, 
+      name=pipeline_version_name, 
+      pipelineid=pipeline_id
+    )
+
+    if self._is_ipython():
+      import IPython
+      html = '<a href=%s/#/pipelines/details/%s>Pipeline details</a>.' % (self._get_url_prefix(), response.id)
       IPython.display.display(IPython.display.HTML(html))
     return response
 
   def get_pipeline(self, pipeline_id):
     """Get pipeline details.
+
     Args:
-      id of the pipeline.
+      pipeline_id: id of the pipeline.
+
     Returns:
       A response object including details of a pipeline.
+
     Throws:
       Exception if pipeline is not found.
     """
@@ -701,12 +936,29 @@ class Client(object):
 
   def delete_pipeline(self, pipeline_id):
     """Delete pipeline.
+
     Args:
-      id of the pipeline.
+      pipeline_id: id of the pipeline.
+
     Returns:
-      Object. If the method is called asynchronously,
-      returns the request thread.
+      Object. If the method is called asynchronously, returns the request thread.
+
     Throws:
       Exception if pipeline is not found.
     """
     return self._pipelines_api.delete_pipeline(id=pipeline_id)
+
+  def list_pipeline_versions(self, pipeline_id, page_token='', page_size=10, sort_by=''):
+    """Lists pipeline versions.
+
+    Args:
+      pipeline_id: Id of the pipeline to list versions
+      page_token: Token for starting of the page.
+      page_size: Size of the page.
+      sort_by: One of 'field_name', 'field_name des'. For example, 'name des'.
+
+    Returns:
+      A response object including a list of versions and next page token.
+    """
+
+    return self._pipelines_api.list_pipeline_versions(page_token=page_token, page_size=page_size, sort_by=sort_by, resource_key_type=kfp_server_api.models.api_resource_type.ApiResourceType.PIPELINE, resource_key_id=pipeline_id)
